module Binary(runPut, runGet, Get, Put,
    getStorable, putStorable,
    getByteString, putByteString,
    putByteStringSz,
    getByteStringLazy, putByteStringLazy,
    putByteStringLazySz,
    getString, putString,
    Binary(..)) where

import Control.Arrow((***))
import Control.Monad
import Control.Monad.Trans
import Control.Monad.Reader
import Control.Monad.State.Strict
import Foreign.ForeignPtr
import Foreign.Ptr
import Foreign.Storable
import Data.ByteString(ByteString)
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Internal as BSI
import Data.Word
import System.IO.Unsafe

import Data.String.UTF8(UTF8)
import qualified Data.String.UTF8 as UTF8
import qualified Data.ByteString.UTF8 as BSUTF8

class Binary a where
    binPut :: a -> Put ()
    binGet :: Get a

instance Binary Int where
    binPut = putStorable
    binGet = getStorable

instance Binary Bool where
    binPut = putStorable
    binGet = getStorable

instance Binary BSL.ByteString where
    binPut = putByteStringLazy
    binGet = getByteStringLazy

instance Binary a => Binary [a] where
    binPut l = binPut (length l) >> mapM_ binPut l
    binGet = binGet >>= flip replicateM binGet

instance (Binary a, Binary b) => Binary (a, b) where
    binPut (a, b) = binPut a >> binPut b
    binGet = liftM2 (,) binGet binGet

instance Binary a => Binary (Maybe a) where
    binPut Nothing  = binPut False
    binPut (Just a) = binPut True >> binPut a
    binGet = binGet >>= (\x -> if x then liftM Just binGet else return Nothing)

type Put = State (Ptr Word8 -> IO (Ptr Word8), Int)

runPut :: Put () -> ByteString
runPut m = unsafePerformIO $ do
    ptr <- mallocForeignPtrBytes s
    _ <- withForeignPtr ptr f
    return $ BSI.fromForeignPtr ptr 0 s
    where (f, s) = execState m (return, 0)
{-# INLINE runPut #-}
    
{-# SPECIALIZE INLINE putStorable :: Int -> Put () #-}
putStorable :: Storable a => a -> Put ()
putStorable x = modify ((>=> store) *** (+ sizeOf x))
    where store p = poke (castPtr p) x >> return (p `plusPtr` sizeOf x)
{-# INLINE putStorable #-}

putByteStringSz :: ByteString -> Put ()
putByteStringSz x = modify ((>=> store) *** (+ len)) where
    (fp, off, len) = BSI.toForeignPtr x
    store p = withForeignPtr fp $
        \c_fp -> BSI.memcpy p (c_fp `plusPtr` off) (fromIntegral len) 
            >> return (p `plusPtr` len)
{-# INLINE putByteStringSz #-}

putByteString :: ByteString -> Put ()
putByteString s = putStorable (BS.length s) >> putByteStringSz s
{-# INLINE putByteString #-}

putByteStringLazySz :: BSL.ByteString -> Put ()
putByteStringLazySz s = 
    mapM_ putByteStringSz (BSL.toChunks s)
{-# INLINE putByteStringLazySz #-}

putByteStringLazy :: BSL.ByteString -> Put ()
putByteStringLazy s = 
    putStorable (fromIntegral $ BSL.length s :: Int) >> putByteStringLazySz s
{-# INLINE putByteStringLazy #-}

putString :: String -> Put ()
putString = putByteString . BSUTF8.fromString

type Get = ReaderT (ForeignPtr Word8, Ptr Word8) (StateT Int IO)

runGet :: ByteString -> Get a -> a
runGet bs m = BSI.inlinePerformIO $ withForeignPtr fp $ 
        \c_fp -> evalStateT (runReaderT m (fp, c_fp)) off
    where (fp, off, len) = BSI.toForeignPtr bs
{-# INLINE runGet #-}

{-# SPECIALIZE getStorable :: Get Word8 #-}
{-# SPECIALIZE getStorable :: Get Int #-}
getStorable :: Storable a => Get a
getStorable = do
    (fp, p) <- ask
    off <- lift get
    r <- liftIO $ peek (castPtr (p `plusPtr` off))
    lift $ put (off + sizeOf r)
    return r
{-# INLINE getStorable #-}

{-
getByteStringSz :: Int -> Get ByteString
getByteStringSz k = do
    (fp, p, off) <- get
    put (fp, p, off + k)
    fp <- liftIO $ mallocForeignPtrBytes k
    liftIO $ withForeignPtr fp $ \fp_c -> BSI.memcpy fp_c (p `plusPtr` off) (fromIntegral k)
    return $ BSI.fromForeignPtr fp 0 k
-}

getByteStringSz :: Int -> Get ByteString
getByteStringSz k = do
    (fp, p) <- ask
    off <- lift get
    lift $ put (off + k)
    return $ BSI.fromForeignPtr fp off k

getByteString :: Get ByteString
getByteString = do
    l <- getStorable
    getByteStringSz l

getByteStringLazy :: Get BSL.ByteString
getByteStringLazy = liftM (BSL.fromChunks . return) getByteString

getString :: Get String
getString = liftM BSUTF8.toString getByteString

