module DatabaseStatistics(TableStats(..),saveStats,loadStats) where

import Database.Berkeley.DbMMa

import Binary
import XmlDatabase

import Control.Monad
import Control.Arrow
import Data.Maybe
import Data.Map(Map)
import qualified Data.Map as M
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL

type Hist = Map DbResult Int

data TableStats = TableStats { 
    statRows :: Int,
    statDifVals :: [Int],
    statHists :: [Maybe Hist]
} deriving Show

instance (Ord k, Binary k, Binary v) => Binary (Map k v) where
    binPut = binPut . M.toList
    binGet = liftM M.fromList binGet

instance Binary TableStats where
    binPut (TableStats r d h) = binPut r >> binPut d >> binPut h
    binGet = liftM3 TableStats binGet binGet binGet

countDifferent :: Db -> IO (Int, Int)
countDifferent db = do
    dbc <- db_cursor [] db Nothing
    calc dbc BS.empty 0 0 
    where
    calc dbc ok rows vals = rows `seq` vals `seq` do
        r <- dbc_move_get [] dbc DB_NEXT
        case r of
            Just (k,v) | ok == k -> calc dbc ok (rows+1) vals
                       | otherwise -> calc dbc k (rows+1) (vals+1)
            Nothing -> return (rows, vals)

navHistograms :: Db -> IO (Hist, Hist)
navHistograms db = do
    dbc <- db_cursor [] db Nothing
    calc dbc M.empty M.empty
    where
    mapinc :: (Num a, Ord k) => k -> Map k a -> Map k a
    mapinc = M.alter (Just . maybe 1 (+1))
    tohist :: (Integral k) => Map k Int -> Hist
    tohist = M.fromAscList . map ((DbInt . fromIntegral) *** id) . M.toList
    calc dbc typs tags = do
        r <- dbc_move_get [] dbc DB_NEXT
        case r of
            Just (kt, nt) -> do
                let k = decodeKey kt
                let NavData typ size tag prn = decodeNav nt
                calc dbc (mapinc typ typs) (mapinc tag tags)
            Nothing -> return (tohist typs, tohist tags)

textHistogram :: Db -> IO Hist
textHistogram db = do
    dbc <- db_cursor [] db Nothing
    calc dbc BS.empty 0 []
    where
    calc dbc s k h = k `seq` do
        r <- dbc_move_get [] dbc DB_NEXT
        case r of
            Just (z, v) | pf <- BS.take 2 z, pf /= s -> 
                            calc dbc pf 1 ((DbString (BSL.fromChunks [s]),k):h)
                        | otherwise ->
                            calc dbc s (k+1) h
            Nothing -> return $ M.fromAscList $ reverse h

calcStats xdbs = do
    (trows, tvals) <- countDifferent (txtIdx xdbs)
    texts <- textHistogram (txtIdx xdbs)
    (prows, pvals) <- countDifferent (parentIdx xdbs)
    (typs, tags) <- navHistograms (navDb xdbs)
    return [("nav", TableStats (prows+1) 
                        [prows+1, pvals, prows+1, M.size typs, M.size tags] 
                        [Nothing, Nothing, Nothing, Just typs, Just tags]),
            ("txt", TableStats (prows+1)
                        [prows+1, ceiling $ fromIntegral tvals / 
                            fromIntegral (prows+1-snd (M.findMin texts))*
                            fromIntegral (prows+1)]
                        [Nothing, Just texts]),
            ("tag", TableStats (M.size tags)
                        [M.size tags, M.size tags]
                        [Nothing, Nothing])]

saveStats path xdbs = do
    s <- calcStats xdbs
    forM_ s $ \(n,ts) -> do
        BS.writeFile (path ++ "/" ++ n ++ ".stat") (runPut $ binPut ts)

loadStats :: String -> IO [(String, TableStats)]
loadStats path = do
    forM ["nav","txt","tag"] $ \n -> do
        liftM (const n &&& flip runGet binGet) $ 
            BS.readFile (path ++ "/" ++ n ++ ".stat")

