module RelationAlgebraStatistics where

import XmlDatabase
import RelationAlgebra
import DatabaseStatistics

import Control.Monad
import Control.Monad.Reader
import qualified Data.Map as Map
import qualified Data.ByteString.Lazy as BSL
import Debug.Trace

type WithStats = Reader [(String, TableStats)]

runWithStats = runReader

selectivity :: RelBExpr -> RelExpr -> WithStats Double
selectivity [] _ = return 1
selectivity (h:t) e = liftM2 (*) (sel1 h) (selectivity t e) where
    sel1 [] = return 0
    sel1 (h:t) = liftM2 max (sel2 h) (sel1 t)
    sel2 (CompB EqC (ColumnT k) (ResultT v)) = do
        s <- ask
        let z = do
            (t, c) <- colsources e !! k
            r <- lookup t s
            hist <- statHists r !! c
            vals <- case v of
                DbInt _ -> Map.lookup v hist
                DbString s -> Map.lookup (DbString $ BSL.take 2 s) hist
            return $ fromIntegral vals / fromIntegral (statRows r)
        case z of
            Nothing -> return 0.7
            Just p -> return p
    sel2 (CompB EqC (ColumnT k) _) = do
        s <- ask
        let z = do
            (t, c) <- colsources e !! k
            r <- lookup t s
            return $ fromIntegral (statDifVals r !! c) / fromIntegral (statRows r)
        case z of
            Nothing -> return 0.7
            Just p -> return p
    sel2 (CompB EqC t (ColumnT k)) = sel2 $ CompB EqC (ColumnT k) t
    sel2 _ = return 0.7

opcount_complex :: (RelExpr -> WithStats Double) -> RelExpr -> WithStats Double
opcount_complex cf (Rename _ _ _ e) = cf e
opcount_complex cf (Filter _ (SimpleFilter _) e) = cf e
opcount_complex cf (Project _ _ e) = cf e
opcount_complex cf (Join _ (NestedLoopJoin b) e1 e2) = do
    c1 <- cf e1
    c2 <- opcount_execute e2
    return $ c1 * c2
opcount_complex cf (Join _ (IndexLoopJoin i q) e1 e2) = do
    c1 <- cf e1
    c2 <- opcount_idxlookup (renumber (+(-num_columns e1)) i) (fmap opaquifyTerm q) e2
    return $ c1 * c2
opcount_complex cf (Limit _ k e) = do
    r <- rowcount e
    c <- cf e
    return $ c / fromIntegral r * fromIntegral k

opcount_execute :: RelExpr -> WithStats Double
opcount_execute r@(RawTable _ _ _) = do
    c <- rowcount r
    return $ fromIntegral c
opcount_execute (ConstTable _ _ _ _ _) = return 1
opcount_execute (Filter _ (IndexFilter i q) e) = 
    opcount_idxlookup i (fmap opaquifyTerm q) e
opcount_execute e = opcount_complex opcount_execute e

opcount_idxlookup :: [Int] -> IdxQuery RelTerm -> RelExpr -> WithStats Double
opcount_idxlookup i q (Project _ s e) = 
    opcount_idxlookup (map (s!!) i) q e
opcount_idxlookup i q (Filter _ (IndexFilter i' q') e) = do
    let b = idxQueryToBExpr i' (fmap opaquifyTerm q)
    opcount_idxlookup i q (filterR (SimpleFilter b) e)
opcount_idxlookup i@[0] (QSingle _) (RawTable _ n _) 
    | n `elem` ["nav","txt","tag"] = return 1
opcount_idxlookup i q e@(RawTable _ _ _) = do
    s <- selectivity (idxQueryToBExpr i q) e
    c <- rowcount e
    return $ fromIntegral c * s
opcount_idxlookup i q e = opcount_complex (opcount_idxlookup i q) e

rowcount :: RelExpr -> WithStats Int
rowcount (RawTable _ t _) = do
    s <- ask
    case lookup t s of
        Just r -> return (statRows r)
        Nothing -> return 0
rowcount (Join _ (NestedLoopJoin []) e1 e2) =
    liftM2 (*) (rowcount e1) (rowcount e2)
rowcount (Join _ (NestedLoopJoin b) e1 e2) = 
    rowcount (filterR (SimpleFilter b) (joinR (NestedLoopJoin []) e1 e2))
rowcount (Join _ (IndexLoopJoin i q) e1 e2) = do
    let b = idxQueryToBExpr i (fmap opaquifyTerm q)
    rowcount (filterR (SimpleFilter b) (joinR (NestedLoopJoin []) e1 e2))
rowcount (Filter _ (SimpleFilter b) e) = 
    liftM ceiling $ liftM2 (*) (selectivity b e) (liftM fromIntegral $ rowcount e)
rowcount (Filter _ (IndexFilter i q) e) = do
    let b = idxQueryToBExpr i (fmap opaquifyTerm q)
    rowcount (filterR (SimpleFilter b) e)
rowcount (ConstTable _ _ _ t _) = return $ length t
rowcount (Rename _ _ _ e) = rowcount e
rowcount (Project _ _ e) = rowcount e
rowcount (Limit _ k e) = return k

opaquifyTerm (ResultT r) = ResultT r
opaquifyTerm (ColumnT c) = ColumnT c
opaquifyTerm (VarT _) = VarT ""
opaquifyTerm (ArithT a t1 t2) =
    case (opaquifyTerm t1, opaquifyTerm t2) of
        (ResultT (DbInt i1), ResultT (DbInt i2)) -> 
            ResultT $ DbInt $ evalArith a i1 i2
        _ -> VarT ""

