module RelationAlgebraVisualize(relexpr_to_dot) where

import RelationAlgebra
import XmlDatabase
import Control.Arrow
import Control.Monad
import Control.Monad.State
import Data.List(intersperse)
import qualified Data.ByteString.Lazy.Char8 as BSL

relexpr_to_dot :: RelExpr -> String
relexpr_to_dot e = snd (execState p (0, id)) "" where
    p = appendLine "digraph {" >> relexpr_h e >> appendLine "}"

newNodeId = do
    i <- liftM fst get
    modify (const (i+1) *** id)
    return ("n" ++ show i)

appendLine l = modify (id *** (. showString "\n") . (. showString l))

put_node e t = do
    i <- newNodeId
    appendLine (i ++ "[shape=record][label=\"{" ++ t ++ "|{"++(display_h "∅" "|" $ freevars e)++"}}\"];")
    return i

put_connection i1 i2 = appendLine (i1 ++ " -> " ++ i2 ++ ";")

tuples_h :: [Tuple] -> String
tuples_h [] = ""
tuples_h (t:ts) = "|{" ++ values_h t ++ "}" ++ tuples_h ts

termtuples_h [] = ""
termtuples_h (t:ts) = "|{" ++ (display_h "∅" "|" . map (relterm_h []) $ t) ++ "}" ++ termtuples_h ts

values_h :: Tuple -> String
values_h = display_h "∅" "|" . map h where
    h (DbInt i) = show i
    h (DbString s) = bsl_h s

binary_node e n e1 e2 = do
    i <- put_node e n
    pi1 <- relexpr_h e1
    pi2 <- relexpr_h e2
    put_connection i pi1
    put_connection i pi2
    return i

unary_node e n e' = do
    i <- put_node e n
    pi <- relexpr_h e'
    put_connection i pi
    return i

relexpr_h e@(RawTable _ name _) = put_node e ("Raw: " ++ name)
relexpr_h e@(ConstTable _ _ _ tuples _) = put_node e ("Const: " ++ termtuples_h tuples)
--relexpr_h e@(Calculate _ _ _ _ e') = unary_node e "Calculate" e'
relexpr_h e@(Limit _ k e') = do
    unary_node e ("Limit " ++ show k) e'
relexpr_h e@(Rename _ _ _ e') = do
    unary_node e ("ρ") e'
relexpr_h e@(Project _ _ e') = do
    unary_node e ("π") e'
relexpr_h e@(Join _ (NestedLoopJoin []) e1 e2) = 
    binary_node e "×" e1 e2
relexpr_h e@(Join _ (NestedLoopJoin b) e1 e2) = 
    binary_node e ("⋈|Nested Loop|" ++ relbexpr_h (freevars e) b) e1 e2
relexpr_h e@(Join _ (IndexLoopJoin i iq) e1 e2) = 
    binary_node e ("⋈|Index Loop|" ++ display_h "" "," (map (freevars e !!) i) ++ idxquery_h (freevars e1) iq) e1 e2
--relexpr_h e@(Union e1 e2) = 
--    binary_node e "∪" e1 e2
relexpr_h e@(Filter _ (SimpleFilter b) e') = do
    unary_node e ("σ|" ++ relbexpr_h (freevars e) b) e'
relexpr_h e@(Filter _ (IndexFilter i iq) e') = do
    unary_node e ("σ|" ++ display_h "" "," (map (freevars e' !!) i) ++ idxquery_h [] iq) e'

relbexpr_h fv = display_h "T" "∧" . map (display_h "F" "∨" . map (relbatom_h fv))

idxquery_h fv (QSingle l) = " = " ++ display_h "" "," (map (relterm_h fv) l)
idxquery_h fv (QRange (Just r1) (Just r2)) = " ∈ (" ++ display_h "" "," (map (relterm_h fv) r1) ++ " ; " ++ display_h "" "," (map (relterm_h fv) r2) ++ ")"

display_h e i [] = e
display_h e i l = concat . intersperse i $ l

comp_h EqC = "="
comp_h NEC = "≠"
comp_h LtC = "&lt;"
comp_h LEC = "≤"
comp_h GtC = "&gt;"
comp_h GEC = "≥"

relbatom_h fv (CompB comp t1 t2) = relterm_h fv t1 ++ comp_h comp ++ relterm_h fv t2

arith_h AddA = "+"
arith_h SubA = "-"
arith_h MulA = "*"

relterm_h fv (ResultT (DbString s)) = bsl_h s
relterm_h fv (ResultT (DbInt s)) = show s
relterm_h fv (ColumnT k) = fv!!k
relterm_h fv (VarT s) = s
relterm_h fv (ArithT a t1 t2) = relterm_h fv t1 ++ arith_h a ++ relterm_h fv t2

bsl_h s = "\\\"" ++ escape (BSL.unpack s) ++ "\\\"" where
    escape = foldr f ""
    f '"' r = "\\\"" ++ r
    f '\\' r = "\\\\" ++ r
    f '<' r = "&lt;" ++ r
    f '>' r = "&gt;" ++ r
    f c r = c:r

