module XQuery(Expr, Module(..), ExprSingle(..), BoolOp(..), CompOp(..),
              CompType(..), ArithOp(..), SetOp(..), UnaryOp(..), StepExpr(..),
              Axis(..), NodeTest(..), Kind(..), Predicate, Quantifier(..),
              ForLetClause(..), WhereClause, OrderByClause, VarBinding(..),
              OrderSpec, OrderModifier, VarName, QName,
              TypeDeclaration, PositionalVar, Attrib(..),
              Prolog(..), PrologItem1, PrologItem2(..),
              doParseModule, xqfreevars) where

import Data.Set(Set)
import qualified Data.Set as Set
import Control.Monad(liftM)
import Data.Maybe(fromMaybe,isJust,maybe)
import Text.XML.HaXml.XmlContent hiding (Prolog)
--import Text.ParserCombinators.Poly

type Expr a = [ExprSingle a]

data Module a = Module (Prolog a) (Expr a) deriving Show

data Prolog a = Prolog [PrologItem1] [PrologItem2 a] deriving Show

type PrologItem1 = ()

data PrologItem2 a = VarDecl VarBinding (Maybe (ExprSingle a)) deriving Show

data ExprSingle a = UserExpr a
                | FLWORExpr [ForLetClause a] (Maybe (WhereClause a)) (OrderByClause a) (ExprSingle a)
                | SequenceExpr (Expr a)
                | ElemConsExpr QName [Attrib a] (Expr a)
                | IfExpr (ExprSingle a) (ExprSingle a) (ExprSingle a)
                | BoolExpr BoolOp (ExprSingle a) (ExprSingle a)
                | CompExpr CompOp (ExprSingle a) (ExprSingle a)
                | RangeExpr (ExprSingle a) (ExprSingle a)
                | ArithExpr ArithOp (ExprSingle a) (ExprSingle a)
                | SetExpr SetOp (ExprSingle a) (ExprSingle a)
                | UnaryExpr UnaryOp (ExprSingle a)
                | PathExpr Bool [StepExpr a]
                | IntegerLiteral Integer
                | DecimalLiteral Rational
                | DoubleLiteral Double
                | StringLiteral String
                | VarRefExpr VarName
                | ContextItemExpr
                | FunctionCallExpr QName [ExprSingle a]
                | OrderedExpr (ExprSingle a)
                | UnorderedExpr (ExprSingle a)
                | QuantifiedExpr Quantifier [(VarBinding, ExprSingle a)] (ExprSingle a)
                deriving Show

data BoolOp = AndOp | OrOp deriving Show

data CompOp = GeneralComp CompType
            | ValueComp CompType
            | IsComp deriving Show

data CompType = EqComp | NeqComp | LtComp | LeqComp | GtComp | GeqComp deriving Show

data ArithOp = AddOp | SubOp | MulOp | DivOp | IdivOp | ModOp deriving Show

data SetOp = UnionOp | IntersectOp | ExceptOp deriving Show

data UnaryOp = UMinusOp | UPlusOp deriving Show

data StepExpr a = AxisStep Axis NodeTest [Predicate a]
                | FilterStep (ExprSingle a) [Predicate a] deriving Show

data Axis = ChildAxis | DescendantAxis | DescendantOrSelfAxis | AttributeAxis deriving Show

data NodeTest = NameTest QName 
              | Wildcard -- TODO, rodzaje wildcardow
              | KindTest Kind deriving Show

data Kind = DocumentKind | ElementKind | AttributeKind | SchemaElementKind
          | SchemaAttributeKind | PIKind | CommentKind | TextKind | AnyKind deriving Show

type Predicate a = Expr a

data VarBinding = VarBinding VarName (Maybe TypeDeclaration) deriving Show

data ForLetClause a = ForClause [(VarBinding, Maybe PositionalVar, ExprSingle a)] deriving Show

type WhereClause = ExprSingle 

type OrderByClause a = [OrderSpec a]

type OrderSpec a = (ExprSingle a, OrderModifier)

type OrderModifier = ()

type VarName = QName

type QName = String

type TypeDeclaration = ()

type PositionalVar = ()

data Attrib a = NsDecl (Maybe String) String
              | AttrCons String (Either (Expr a) String) deriving Show

data Quantifier = SomeQ | EveryQ deriving Show

instance HTypeable (ExprSingle a) where
    toHType _ = Defined "ExprSingle" [] []

instance HTypeable (ForLetClause a) where
    toHType _ = Defined "ForLetClause" [] []

instance HTypeable (StepExpr a) where
    toHType _ = Defined "StepExpr" [] []

instance HTypeable Kind where
    toHType _ = Defined "Kind" [] []

instance HTypeable Quantifier where
    toHType _ = Defined "Quantifier" [] []

instance HTypeable (Module a) where
    toHType _ = Defined "Module" [] []

instance HTypeable (Prolog a) where
    toHType _ = Defined "Prolog" [] []

instance HTypeable (PrologItem2 a) where
    toHType _ = Defined "PrologItem2" [] []

instance XmlContent (ExprSingle a) where
    parseContents = oneOf [parseFlwor, parsePath, parseFunctionCallExpr,
        parseUnaryOp, parseBinaryOp, parseElementConstructor,
        parseIntegerConstant, parseDoubleConstant, parseStringConstant,
        parseVarRef, parseContextItem, parseWrappers, parseSequenceExpr,
        parseIfExpr, parseQuantifiedExpr] 

instance XmlContent (ForLetClause a) where
    parseContents = parseForLet

instance XmlContent (StepExpr a) where
    parseContents = parseStepExpr

instance XmlContent Kind where
    parseContents = parseKind

instance XmlContent (Module a) where
    parseContents = parseModule

instance XmlContent Quantifier where
    parseContents = parseQuantifier

instance XmlContent (Prolog a) where
    parseContents = parseProlog

instance XmlContent (PrologItem2 a) where
    parseContents = parsePrologItem2

parseExprList :: XMLParser (Expr a)
parseExprList = many parseContents

parsePredicates = inElement "xqx:predicates" parseExprList

parseExprsH n p = do 
    e <- element n
    commit (interior e (oneOf [p e, parseContents]))

parseExprH n = parseExprsH [n]

parseTypedVariableBinding = inElement "xqx:typedVariableBinding" $ do
    vn <- inElement "xqx:varName" text
    optional (element ["xqx:typeDeclaration"])
    return (VarBinding vn Nothing)

parseForLet = inElement "xqx:forClause" (liftM ForClause $ many1 forClauseItem) where
    forClauseItem = inElement "xqx:forClauseItem" $ do
        vb <- parseTypedVariableBinding
        optional (element ["xqx:positionalVariableBinding"])
        -- exts
        expr <- inElement "xqx:forExpr" parseContents
        return (vb, Nothing, expr)

parseFlwor = parseExprH "xqx:flworExpr" inFlwor where
    inFlwor _ = do
        forLetCl <- many1 parseContents
        whereCl <- optional (inElement "xqx:whereClause" parseContents)
        optional (element ["xqx:orderByClause"])
        -- orderByCl
        let orderByCl = []
        returnCl <- inElement "xqx:returnClause" parseContents
        return (FLWORExpr forLetCl whereCl [] returnCl)

parseIntegerConstant = 
    parseConstantH "xqx:integerConstantExpr" (IntegerLiteral . read)

parseDoubleConstant = 
    parseConstantH "xqx:doubleConstantExpr" (DoubleLiteral . read)

parseStringConstant = parseConstantH "xqx:stringConstantExpr" StringLiteral

parseConstantH c t = parseExprH c inConstant where
    inConstant _ = liftM t $ inElement "xqx:value" text

parseVarRef = parseExprH "xqx:varRef" inVarRef where
    inVarRef _ = liftM VarRefExpr $ inElement "xqx:name" text

parseWrappers :: XMLParser (ExprSingle a)
parseWrappers = parseExprsH wrappers (\_ -> fail "wrap") where
    wrappers = []

parseContextItem = parseExprH "xqx:contextItemExpr" inCtx where
    inCtx _ = return ContextItemExpr

parseUnaryOp = parseExprsH [fst op | op <- unaryops] inOp where
    unaryops = [("xqx:unaryPlusOp", UPlusOp), ("xqx:unaryMinusOp", UMinusOp)]
    inOp (Elem n _ _) = liftM (UnaryExpr (lookupU n unaryops)) $ 
            inElement "xqx:operand" parseContents 

parseBinaryOp = parseExprsH [fst op | op <- binaryops] inOp where
    binaryops = [("xqx:addOp",      ArithExpr AddOp), 
                 ("xqx:subtractOp", ArithExpr SubOp),
                 ("xqx:multiplyOp", ArithExpr MulOp),
                 ("xqx:divOp",      ArithExpr DivOp),
                 ("xqx:idivOp",     ArithExpr IdivOp),
                 ("xqx:modOp",      ArithExpr ModOp),
                 ("xqx:isOp", CompExpr IsComp),
                 ("xqx:eqOp", CompExpr (ValueComp EqComp)),
                 ("xqx:neOp", CompExpr (ValueComp NeqComp)),
                 ("xqx:gtOp", CompExpr (ValueComp GtComp)),
                 ("xqx:geOp", CompExpr (ValueComp GeqComp)),
                 ("xqx:ltOp", CompExpr (ValueComp LtComp)),
                 ("xqx:leOp", CompExpr (ValueComp LeqComp)),
                 ("xqx:equalOp",              CompExpr (GeneralComp EqComp)),
                 ("xqx:notEqualOp",           CompExpr (GeneralComp NeqComp)),
                 ("xqx:greaterThanOp",        CompExpr (GeneralComp GtComp)),
                 ("xqx:greaterThanOrEqualOp", CompExpr (GeneralComp GeqComp)),
                 ("xqx:lessThanOp",           CompExpr (GeneralComp LtComp)),
                 ("xqx:lessThanOrEqualOp",    CompExpr (GeneralComp LeqComp)),
                 ("xqx:andOp", BoolExpr AndOp),
                 ("xqx:orOp",  BoolExpr OrOp),
                 ("xqx:unionOp",     SetExpr UnionOp),
                 ("xqx:intersectOp", SetExpr IntersectOp),
                 ("xqx:exceptOp",    SetExpr ExceptOp)]
    inOp (Elem n _ _) = do 
        e1 <- inElement "xqx:firstOperand" parseContents 
        e2 <- inElement "xqx:secondOperand" parseContents 
        return (lookupU n binaryops e1 e2)

parsePath = parseExprH "xqx:pathExpr" inPath where
    inPath _ = do
        i <- liftM isJust $ optional (inElement "xqx:rootExpr" $ return ())
        k <- many parseContents
        return (PathExpr i (fixPath k)) -- usuwanie descendantOrSelf

lookupU n t = fromMaybe undefined $ lookup n t

parseStepExpr = oneOf [inElement "xqx:argExpr" inArg, inElement "xqx:stepExpr" (oneOf [inAxis, inFilter])] where
    axes = [("child", ChildAxis), 
            ("descendant", DescendantAxis),
            ("descendant-or-self", DescendantOrSelfAxis),
            ("attribute", AttributeAxis)]
    nameTest = liftM NameTest $ inElement "xqx:nameTest" text
    kindTest = liftM KindTest $ parseContents
    wildcard = inElement "xqx:Wildcard" $ do
        many next
        return Wildcard
    inAxis = do
        axis <- inElement "xqx:xpathAxis" text
        nodetest <- oneOf [nameTest, kindTest, wildcard]
        preds <- many parsePredicates
        return (AxisStep (lookupU axis axes) nodetest preds)
    inFilter = do 
        fexpr <- inElement "xqx:filterExpr" parseContents
        preds <- many parsePredicates
        return (FilterStep fexpr preds)
    -- stary xqueryx
    inArg = do
        fexpr <- parseContents
        return (FilterStep fexpr [])

parseKind = element [fst k | k <- kinds] >>= inKind where
    kinds = [("xqx:anyKindTest", AnyKind),
             ("xqx:textTest", TextKind)]
    inKind (Elem n _ _) = return (lookupU n kinds)

parseElementConstructor = parseExprH "xqx:elementConstructor" inEC where
    inEC _ = do
        name <- inElement "xqx:tagName" text
        attrs <- optional $ parseAttributeList
        cnt <- optional $ inElement "xqx:elementContent" parseExprList
        return (ElemConsExpr name (maybe [] id attrs) (maybe [] id cnt))

parseAttributeList = inElement "xqx:attributeList" inAL where
    inAL = many (oneOf [attributeConstructor, namespaceDeclaration])
    attributeConstructor = inElement "xqx:attributeConstructor" $ do
        name <- inElement "xqx:attributeName" text
        value <- oneOf [inElement "xqx:attributeValueExpr" $ liftM Left parseExprList,
            inElement "xqx:attributeValue" $ liftM Right text]
        return (AttrCons name value)
    namespaceDeclaration = inElement "xqx:namespaceDeclaration" $ do
        prefix <- optional $ inElement "xqx:prefix" text
        uri <- inElement "xqx:uri" text
        return (NsDecl prefix uri)

parseFunctionCallExpr = parseExprH "xqx:functionCallExpr" inFC where
    inFC _ = do
        fun <- inElement "xqx:functionName" text
        args <- optional $ inElement "xqx:arguments" (many parseContents)
        return (FunctionCallExpr fun (maybe [] id args))

parseIfExpr = parseExprH "xqx:ifThenElseExpr" inIf where
    inIf _ = do
        ifc <- inElement "xqx:ifClause" parseContents
        thenc <- inElement "xqx:thenClause" parseContents
        elsec <- inElement "xqx:elseClause" parseContents
        return (IfExpr ifc thenc elsec)

parseQuantifiedExpr = parseExprH "xqx:quantifiedExpr" inQE where
    inQE _ = do
        qf <- parseContents
        qexpr <- many1 $ inElement "xqx:quantifiedExprInClause" $ do
            tv <- parseTypedVariableBinding
            e <- inElement "xqx:sourceExpr" parseContents
            return (tv, e)
        pred <- inElement "xqx:predicateExpr" parseContents
        return (QuantifiedExpr qf qexpr pred)

parseQuantifier = inElement "xqx:quantifier" inQ where
    inQ = do
        t <- text
        case t of
            "some" -> return SomeQ
            "every" -> return EveryQ

parseSequenceExpr = parseExprH "xqx:sequenceExpr" inSeqE where
    inSeqE _ = liftM SequenceExpr parseExprList

parseProlog = inElement "xqx:prolog" inProlog where
    inProlog = do
        it2 <- many parseContents
        return (Prolog [] it2)

parsePrologItem2 = oneOf [varDecl] where
    varDecl = inElement "xqx:varDecl" $ do
        vn <- inElement "xqx:varName" text
        optional (element ["xqx:typeDeclaration"])
        val <- oneOf [
            inElement "xqx:varValue" $ liftM Just parseContents,
            inElement "xqx:external" $ return Nothing]
        return (VarDecl (VarBinding vn Nothing) val)

parseModule = inElement "xqx:module" $ inElement "xqx:mainModule" mmod where
    mmod = do
        prolog <- optional parseContents
        body <- inElement "xqx:queryBody" (many parseContents)
        return (Module (maybe (Prolog [] []) id prolog) body)

doParseExpr :: String -> Either String (ExprSingle a)
doParseExpr = readXml

doParseModule :: String -> Either String (Module a)
doParseModule = readXml 

fixPath (AxisStep DescendantOrSelfAxis (KindTest AnyKind) []:AxisStep ChildAxis nt []:ass) = fixPath (AxisStep DescendantAxis nt []:ass)
fixPath (as:ass) = as:fixPath ass
fixPath [] = []

xqfreevars (FLWORExpr v _ _ e) = (xqfreevars e `Set.union` Set.unions (map (\(ForClause l) -> Set.unions $ map (\(_,_,a) -> xqfreevars a) l) v)) `Set.difference` (Set.unions $ map (\(ForClause l) -> Set.fromList $ map (\(VarBinding a _,_,_) -> a) l) v)
xqfreevars (QuantifiedExpr _ v e) = (xqfreevars e `Set.union` Set.unions (map (xqfreevars . snd) v)) `Set.difference` (Set.fromList $ map ((\(VarBinding a _) -> a) . fst) v)
xqfreevars (SequenceExpr es) = Set.unions $ map xqfreevars es
xqfreevars (ElemConsExpr _ _ es) = Set.unions $ map xqfreevars es
xqfreevars (IfExpr e1 e2 e3) = Set.unions $ map xqfreevars [e1,e2,e3]
xqfreevars (BoolExpr _ e1 e2) = xqfreevars e1 `Set.union` xqfreevars e2
xqfreevars (CompExpr _ e1 e2) = xqfreevars e1 `Set.union` xqfreevars e2
xqfreevars (UnaryExpr _ e) = xqfreevars e
xqfreevars (FunctionCallExpr _ es) = Set.unions $ map xqfreevars es
xqfreevars (PathExpr _ ps) = Set.unions $ map pfreevars ps where
    pfreevars (FilterStep e _) = xqfreevars e
    pfreevars _ = Set.empty
xqfreevars _ = Set.empty

