module XQuerySemantics where

import qualified XQuery as XQ
import qualified Data.Map as Map
import Data.Maybe(fromJust)
import XmlBS
--import Text.XML.HaXml.Posn
import MonadExponent
import MonadWithList
import XmlDatabase
import XQuery
import RelationAlgebra hiding (Subst)
import Control.Arrow
import Control.Monad
import Control.Monad.Trans
import Control.Monad.Reader
import qualified Data.ByteString as BS
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Lazy.Char8 as BSLC8
import LazyIO

import Debug.Trace

data RelQuery = IterN [XQ.VarName] [Int] RelBExpr Int (ExprSingle RelQuery)
              | Iter [XQ.VarName] RelExpr (ExprSingle RelQuery) 
    deriving Show

data TypedContent = DocumentC (Document Mark) 
                  | ContentC (Content Mark) 
                  | BooleanC Bool
                  | StringC BSL.ByteString
                  | UntypedAtomicC BSL.ByteString 

instance Show TypedContent where
    show (DocumentC d) = "doc"
    show (ContentC c) = "content"
    show (BooleanC b) = "boolean " ++ show b
    show (StringC s) = "string " ++ show s
    show (UntypedAtomicC s) = "untyped " ++ show s

type Subst = Map.Map XQ.VarName TypedContent

type Sem = ExponentT (Subst, Xdbs) LazyIO

contentToString :: [TypedContent] -> BSL.ByteString
contentToString = BSL.concat . map ts where
    ts (ContentC c) = ts (typedValue c)
    ts (BooleanC True) = BSLC8.pack "true"
    ts (BooleanC False) = BSLC8.pack "false"
    ts (StringC s) = s
    ts (UntypedAtomicC s) = s

contentToBoolean :: [TypedContent] -> Bool
contentToBoolean [] = False
contentToBoolean (ContentC (Elem (CElem _ _ _) _):_) = True
contentToBoolean [BooleanC b] = b
contentToBoolean [StringC s] = not $ BSL.null s
contentToBoolean [UntypedAtomicC s] = not $ BSL.null s
contentToBoolean _ = True
{-
contentToBoolean x = error ("contentToBoolean: type error: " ++ show x)
-}

atomize :: TypedContent -> TypedContent
atomize (DocumentC (Document _ _ e _)) = typedValue e
atomize (ContentC c) = typedValue c
atomize e = e

entityValue "amp" = "&"
entityValue "quot" = "\""
entityValue "lt" = "<"
entityValue "gt" = ">"
entityValue "apos" = "'"
entityValue _ = ""

typedValue :: Content Mark -> TypedContent
typedValue (Elem (CString _ s) _) = UntypedAtomicC s
typedValue (Elem (CRef (RefEntity s)) _) = UntypedAtomicC $ BSLC8.pack $ entityValue s
typedValue (Elem (CRef (RefChar c)) _) = UntypedAtomicC $ BSL.singleton (fromIntegral c)
typedValue (Elem (CElem _ _ c) _) = 
    UntypedAtomicC (contentToString (map typedValue c))
typedValue (Elem (CMisc (Comment c)) _) = StringC (BSLC8.pack c)
typedValue (Elem (CMisc (PI (_, p))) _) = StringC (BSLC8.pack p)

contentToXML :: TypedContent -> Content Mark
contentToXML (ContentC e) = e
contentToXML (BooleanC True) = Elem (CString False (BSLC8.pack "true")) Generated
contentToXML (BooleanC False) = Elem (CString False (BSLC8.pack "false")) Generated
contentToXML (StringC s) = Elem (CString False s) Generated

evalAttr :: XQ.Attrib RelQuery -> Sem Attribute
evalAttr (XQ.AttrCons t (Right v)) = return (BSLC8.pack t, AttValue [Left $ BSLC8.pack v])

evalExpr :: XQ.Expr RelQuery -> Sem [TypedContent]
evalExpr e = return . concat =<< mapM evalExprSingle e 

evalBoolOp :: XQ.BoolOp -> Bool -> Bool -> Bool
evalBoolOp XQ.AndOp = (&&)
evalBoolOp XQ.OrOp  = (||)

evalCompType :: Ord a => XQ.CompType -> a -> a -> Bool
evalCompType XQ.EqComp  = (==)
evalCompType XQ.NeqComp = (/=)
evalCompType XQ.LtComp  = (<)
evalCompType XQ.LeqComp = (<=)
evalCompType XQ.GtComp  = (>)
evalCompType XQ.GeqComp = (>=)

evalValueCompType :: XQ.CompType -> TypedContent -> TypedContent -> Bool
evalValueCompType op (BooleanC b1) (BooleanC b2) = evalCompType op b1 b2
evalValueCompType op (StringC b1) (StringC b2) = evalCompType op b1 b2
evalValueCompType op (UntypedAtomicC b1) (UntypedAtomicC b2) = 
    evalCompType op b1 b2
evalValueCompType op (UntypedAtomicC s1) (StringC s2) = 
    evalValueCompType op (StringC s1) (StringC s2)
evalValueCompType op a (UntypedAtomicC b) = 
    evalValueCompType op (UntypedAtomicC b) a

evalAxis :: XQ.Axis -> TypedContent -> Sem [TypedContent]
evalAxis XQ.ChildAxis (ContentC e@(Elem _ (PrePost (Just _) _ _))) = do
    d <- liftIO $ retrieveChildren e
    return $ map ContentC d
evalAxis XQ.ChildAxis (DocumentC (Document _ _ e _)) = 
    return [ContentC e]
evalAxis XQ.ChildAxis (ContentC (Elem (CElem _ _ c) _)) = 
    return $ map ContentC c
evalAxis XQ.ChildAxis _ = return []
evalAxis XQ.DescendantOrSelfAxis c = 
    liftM (c:) $ evalAxis XQ.DescendantAxis c
evalAxis XQ.DescendantAxis (ContentC e@(Elem _ (PrePost (Just _) _ _))) = do
    d <- liftIO $ retrieveDescendants e
    return $ map ContentC d
evalAxis XQ.DescendantAxis c = do
    cs <- evalAxis XQ.ChildAxis c
    rs <- mapM (evalAxis XQ.DescendantOrSelfAxis) cs
    return $ concat rs

evalNodeTest :: XQ.NodeTest -> TypedContent -> Bool
evalNodeTest (XQ.NameTest n) (ContentC (Elem (CElem n' _ _) _)) = n == BSLC8.unpack n'
evalNodeTest (XQ.NameTest n) _ = False
evalNodeTest XQ.Wildcard _ = True
evalNodeTest (XQ.KindTest XQ.AnyKind) _ = True
evalNodeTest (XQ.KindTest XQ.TextKind) (ContentC (Elem (CString _ _) _)) = True
evalNodeTest (XQ.KindTest XQ.TextKind) _ = False
evalNodeTest (XQ.KindTest XQ.ElementKind) (ContentC (Elem (CElem _ _ _) _)) = True
evalNodeTest (XQ.KindTest XQ.ElementKind) _ = False

evalStepExpr :: XQ.StepExpr RelQuery -> TypedContent -> Sem [TypedContent]
evalStepExpr (XQ.AxisStep a t []) e = do
    ea <- evalAxis a e
    return $ filter (evalNodeTest t) ea
evalStepExpr (XQ.FilterStep f []) _ = evalExprSingle f

evalQuantifier :: XQ.Quantifier -> [Bool] -> Bool
evalQuantifier XQ.SomeQ = or
evalQuantifier XQ.EveryQ = and

content2rel :: TypedContent -> DbResult
content2rel (DocumentC _) = DbInt 1
content2rel (ContentC (Elem _ (PrePost _ pre _))) = DbInt pre
content2rel (BooleanC b) = DbBool b
content2rel (StringC s) = DbString s
content2rel (UntypedAtomicC s) = DbString s

getSubst = liftM fst getExponent
getDb = liftM snd getExponent
setSubst s = modifySubst (const s)
modifySubst f = modifyExponent (f *** id)

intToContent (DbInt 1) = getDb >>= \db -> liftM (\x -> DocumentC (Document (XmlBS.Prolog (Just (XMLDecl "1.0" Nothing Nothing)) [] Nothing []) [] x [])) $ liftIO $ retrieveContentFromDB db 2
intToContent (DbInt i) = getDb >>= \db -> liftM ContentC $ liftIO $ retrieveContentFromDB db i

evalExprSingle :: XQ.ExprSingle RelQuery -> Sem [TypedContent]
evalExprSingle (XQ.UserExpr (Iter vars rel e)) = do
    (s, db) <- getExponent
    ts <- liftIO $ runLazyIO $ runReaderT (execute rel) (db, Map.map content2rel s)
    r <- forM ts $ \tuple -> do
        tuple' <- forM tuple intToContent
        setSubst (foldr (uncurry Map.insert) s (zip vars tuple')) (evalExprSingle e)
    return $ concat r
evalExprSingle (XQ.ElemConsExpr n as e) = do
    as' <- mapM evalAttr as
    v <- evalExpr e
    return [ContentC $ Elem (CElem (BSLC8.pack n) as' (map contentToXML v)) Generated]
evalExprSingle (XQ.VarRefExpr v) = do 
    s <- getSubst
    return [fromJust $ Map.lookup v s]
evalExprSingle (XQ.BoolExpr op e1 e2) = do
    v1 <- evalExprSingle e1
    v2 <- evalExprSingle e2
    return [BooleanC $ evalBoolOp op
        (contentToBoolean v1) (contentToBoolean v2)]
evalExprSingle (XQ.CompExpr (XQ.GeneralComp op) e1 e2) = do
    v1s <- evalExprSingle e1
    v2s <- evalExprSingle e2
    return [BooleanC $ or [evalValueCompType op (atomize v1) (atomize v2) |
                    v1 <- v1s, v2 <- v2s]]
evalExprSingle (XQ.FunctionCallExpr "true" []) = return [BooleanC True]
evalExprSingle (XQ.FunctionCallExpr "false" []) = return [BooleanC False]
evalExprSingle (XQ.FunctionCallExpr "not" [e]) = do
    v <- evalExprSingle e
    return [BooleanC $ not $ contentToBoolean v]
evalExprSingle (XQ.IfExpr i t e) = do
    v <- evalExprSingle i
    if contentToBoolean v then evalExprSingle t else evalExprSingle e
evalExprSingle (XQ.FLWORExpr [XQ.ForClause [(XQ.VarBinding vn _,_,e)]] Nothing [] r) = do
    v <- evalExprSingle e
    r <- forM v (\e' -> modifySubst (Map.insert vn e') (evalExprSingle r))
    return $ concat r
-- nie ma eliminacji duplikatow, ale to i tak poza naszym fragmentem
evalExprSingle (XQ.PathExpr False ses) = unwithlist $
    foldr (>=>) return [WithList . evalStepExpr se | se <- ses] (error "noctx")
evalExprSingle (XQ.QuantifiedExpr q [(XQ.VarBinding vn _, e)] cond) = do
    v <- evalExprSingle e
    c <- forM v (\e' -> modifySubst (Map.insert vn e') (evalExprSingle cond))
    return [BooleanC $ evalQuantifier q (map contentToBoolean c)]
evalExprSingle (XQ.SequenceExpr e) = evalExpr e
evalExprSingle (XQ.StringLiteral s) = return [StringC (BSLC8.pack s)]
-- liczby sa poza naszym fragmentem
evalExprSingle (XQ.IntegerLiteral s) = return [StringC (BSLC8.pack $ show s)]
evalExprSingle e = error (show e)

