module XmlDatabase(Mark(..), storeInDB, retrieveFromDB, retrieveContentFromDB,
    retrieveDescendants, retrieveChildren, Xdbs(..), NavData(..), DbResult(..),
    encodeKey, decodeKey, encodeNav, decodeNav) where

import Database.Berkeley.DbMMa

import Control.Monad
import Control.Arrow(first)
import qualified Control.Monad.State as CMS
import qualified Control.Monad as CM
import XmlBS

import Foreign.ForeignPtr
import Foreign.Ptr
import Foreign.Storable

import Data.Word
import Data.ByteString(ByteString)
import qualified Data.ByteString.Internal as BSI
import qualified Data.ByteString.Lazy as BSL
import qualified Data.ByteString.Char8 as BSC8
import qualified Data.ByteString as BS

import Debug.Trace

import System.IO.Unsafe

import ByteOrder
import Binary

{-# RULES
"encodeKey/decodeKey" forall x. decodeKey (encodeKey x) = x
"decodeKey/encodeKey" forall x. encodeKey (decodeKey x) = x
"encodeNav/decodeNav" forall x. decodeNav (encodeNav x) = x
"decodeNav/encodeNav" forall x. encodeNav (decodeNav x) = x
  #-}

data Xdbs = Xdbs { navDb :: Db, txtDb :: Db, tagDb :: Db, tagIdx :: Db, parentIdx :: Db, txtIdx :: Db }

instance Show Xdbs where
    show _ = "Xdbs"

data Mark = Generated | PrePost (Maybe Xdbs) Int Int 

-- Oznaczanie Pre/Post.

type MarkerM = CMS.State Int

data DbResult = DbBool Bool | DbInt Int | DbString BSL.ByteString deriving Show

instance Eq DbResult where
    DbInt i1 == DbInt i2 = i1 == i2
    DbString s1 == DbString s2 = s1 == s2
    DbBool i1 == DbBool i2 = i1 == i2

instance Ord DbResult where
    compare (DbInt i1) (DbInt i2) = compare i1 i2
    compare (DbString s1) (DbString s2) = compare s1 s2
    compare (DbBool b1) (DbBool b2) = compare b1 b2

instance Num DbResult where
    DbInt i1 + DbInt i2 = DbInt (i1 + i2)
    DbInt i1 - DbInt i2 = DbInt (i1 - i2)
    DbInt i1 * DbInt i2 = DbInt (i1 * i2)
    negate (DbInt i) = DbInt (negate i)
    abs (DbInt i) = DbInt (abs i)
    signum (DbInt i) = DbInt (signum i)
    fromInteger i = DbInt (fromInteger i)

instance Binary DbResult where
    binPut (DbBool b) = binPut (0 :: Int) >> binPut b
    binPut (DbInt i) = binPut (1 :: Int) >> binPut i
    binPut (DbString s) = binPut (2 :: Int) >> binPut s
    binGet = binGet >>= ([liftM DbBool binGet, liftM DbInt binGet, 
                          liftM DbString binGet] !!)

newPreM :: MarkerM Int
newPreM = CMS.get >>= \v -> CMS.put (v + 1) >> return v

curPreM :: MarkerM Int
curPreM = CMS.get

markPrePost :: Content a -> Content Mark
markPrePost x = CMS.evalState (mpp x) 2 where
    e = Nothing
    mpp (Elem (CElem a b l) _) = do
        pre <- newPreM
        l' <- mapM mpp l
        pre' <- curPreM
        return (Elem (CElem a b l') (PrePost e pre pre'))
    mpp (Elem (CString a b) _) = do
        pre <- newPreM
        return (Elem (CString a b) (PrePost e pre 1))
    mpp (Elem (CRef a) _) = do
        pre <- newPreM
        return (Elem (CRef a) (PrePost e pre 1))
    mpp (Elem (CMisc a) _) = do
        pre <- newPreM
        return (Elem (CMisc a) (PrePost e pre 1))

-- Kodowanie klucza. Big-Endian jest kluczowy dla prawidłowego porównywania
-- w bazie danych.

encodeKey :: Int -> ByteString
encodeKey k = runPut $ putStorable k

decodeKey :: ByteString -> Int
decodeKey b = BSI.inlinePerformIO $ withByteString b $ \c_str str_len -> do
    i <- peekByteOff c_str 0
    return i
{-# INLINE decodeKey #-}

-- Konwersja do ByteString. 

data NavData = NavData Word8 Int Int Int deriving Show

ctHelper :: Word8 -> Int -> Int -> Mark -> ByteString -> (NavData, ByteString)
ctHelper typ tag prn mark stor = (markToNav typ tag prn mark, stor)

markToNav :: Word8 -> Int -> Int -> Mark -> NavData
markToNav typ tag prn (PrePost _ _ size) = NavData typ size tag prn

getAddKey :: Xdbs -> ByteString -> IO Int
getAddKey db s = do
    v <- db_get [] (tagIdx db) Nothing s
    case v of
        Just k -> return (decodeKey k)
        Nothing -> do
            k <- db_put_append [] (tagDb db) Nothing s
            db_put [] (tagIdx db) Nothing s (encodeKey k)
            return k

contentToNav :: Xdbs -> Int -> Content Mark -> IO (NavData, ByteString)
contentToNav db prn (Elem (CElem n _ _) mark) = do
    k <- getAddKey db (BS.concat $ BSL.toChunks n)
    return $ ctHelper 1 k prn mark BS.empty
contentToNav db prn (Elem (CString b c) mark) = 
    return $ ctHelper 2 1 prn mark (BS.concat $ BSL.toChunks c)
contentToNav db prn (Elem (CRef (RefEntity s)) mark) = do
    k <- getAddKey db (BSC8.pack s)
    return $ ctHelper 3 k prn mark (BSC8.pack s)
contentToNav db prn (Elem (CRef (RefChar c)) mark) = 
    return $ ctHelper 4 1 prn mark (runPut $ putStorable c)
contentToNav db prn (Elem (CMisc (Comment s)) mark) = 
    return $ ctHelper 5 1 prn mark (BSC8.pack s)
contentToNav db prn (Elem (CMisc (PI (s1, s2))) mark) = do
    k <- getAddKey db (BSC8.pack s1)
    return $ ctHelper 6 k prn mark (BSC8.pack s2)

encodeNav :: NavData -> ByteString
encodeNav (NavData typ size tag prn) = 
    runPut $ putStorable size >> putStorable prn >> putStorable tag >> putStorable typ

decodeNav :: ByteString -> NavData
decodeNav bs = runGet bs $ liftM4 nd getStorable getStorable getStorable getStorable
    where nd a b c d = NavData d a c b

--contentToVal :: Content Mark -> (ByteString, ByteString)
--contentToVal = first encodeNav . contentToNav

withByteString :: ByteString -> (Ptr Word8 -> Int -> IO a) -> IO a
withByteString bs code = do
    let (fp, fp_offset, length) = BSI.toForeignPtr bs
    withForeignPtr fp $ \c_fp ->
        code (c_fp `plusPtr` fp_offset) length
{-# INLINE withByteString #-}

valToContent :: Xdbs -> Int -> ByteString -> ByteString -> IO (Content Mark)
valToContent db pre str1 str2 = do
    h <- helper (typ :: Word8)
    return $ Elem h (PrePost (Just db) pre size)
    where
    NavData typ size tag _ = decodeNav str1
    helper 1 = do
        ~(Just t) <- unsafeInterleaveIO $ db_get [] (tagDb db) Nothing (encodeKey tag)
        return $ CElem (BSL.fromChunks [t]) [] []
    helper 2 = return $ CString False (BSL.fromChunks [str2])
    helper 3 = do
        ~(Just t) <- unsafeInterleaveIO $ db_get [] (tagDb db) Nothing (encodeKey tag)
        return $ CRef (RefEntity (BSC8.unpack t))
    helper 4 = return $ CRef (RefChar (runGet str2 getStorable))
    helper 5 = return $ CMisc (Comment (BSC8.unpack str2))
    helper 6 = do
        ~(Just t) <- unsafeInterleaveIO $ db_get [] (tagDb db) Nothing (encodeKey tag)
        return $ CMisc (PI (BSC8.unpack t, BSC8.unpack str2))

-- Rekurencyjne zapisywanie danych w bazie, dbamy o dobre zachowanie
-- pamięciowe dla dużych danych.

storeContentInDB db prn c = do
    -- najpierw zapisujemy tylko pre
    let (PrePost _ pre post) = info c
    (v1@(NavData typ size tag _), v2) <- contentToNav db prn c
    db_put [] (parentIdx db) Nothing (encodeKey prn) (encodeKey pre)
    db_put_append [] (navDb db) Nothing (encodeNav $ NavData typ 1 tag prn)
    db_put_append [] (txtDb db) Nothing v2
    db_put [] (txtIdx db) Nothing v2 (encodeKey pre)
    case c of
        Elem (CElem _ _ cs@(_:_)) _ -> do
            mapM_ (storeContentInDB db pre) cs
            -- potem dopisujemy size
            db_put [] (navDb db) Nothing (encodeKey pre) (encodeNav $ NavData typ (post-pre) tag prn)
        _ -> return ()

-- Odczytywanie danych z bazy. Gwarantujemy leniwy odczyt potomków z bazy
-- jednym skanem sekwencyjnym, dzięki czemu można wypisywać w prawie
-- stałej pamięci.

valToContentRec :: Xdbs -> Int -> ByteString -> ByteString -> IO (Content Mark)
valToContentRec db pre v1 v2 = do
    ee <- valToContent db pre v1 v2
    case ee of
        Elem ~(CElem n a _) mark@(PrePost _ pre size) | size > 1 -> do
            c <- unsafeInterleaveIO $ do
                dbc <- db_cursor [] (navDb db) Nothing
                _ <- dbc_get_sz (Just 13) [] dbc (encodeKey $ pre + 1)
                (c, _) <- scanDescendants db dbc DB_CURRENT (pre + size) (pre + size) True
                return c
            return (Elem (CElem n a c) mark)
        ret -> return ret

scanDescendants :: Xdbs -> Dbc -> DbDir -> Int -> Int -> Bool -> IO ([Content Mark], Bool)
scanDescendants db dbc dir end aend False = return ([], False)
scanDescendants db dbc dir end aend True = do
    (k, v1) <- liftM addGuard $ dbc_move_get_sz (Just 4) (Just 13) [] dbc dir
    ~(Just v2) <- unsafeInterleaveIO $ db_get [] (txtDb db) Nothing k
    if decodeKey k >= aend then dbc_close dbc >> return ([], False) else 
        if decodeKey k >= end then return ([], True) else do
    ee <- valToContent db (decodeKey k) v1 v2
    case ee of
        Elem ~(CElem n a _) mark@(PrePost _ pre size) | size > 1 -> do
            ~(chl, sqt') <- unsafeInterleaveIO $
                scanDescendants db dbc DB_NEXT (pre + size) aend True
            ~(cs, sqt'') <- unsafeInterleaveIO $ 
                scanDescendants db dbc DB_CURRENT end aend sqt'
            return (Elem (CElem n a chl) mark:cs, sqt'')
        c -> do
            (cs, sqt') <- unsafeInterleaveIO $ 
                scanDescendants db dbc DB_NEXT end aend True
            return (c:cs, sqt')

retrieveContentFromDB :: Xdbs -> Int -> IO (Content Mark)
retrieveContentFromDB db k = do
    (Just v1) <- db_get [] (navDb db) Nothing (encodeKey k)
    ~(Just v2) <- unsafeInterleaveIO $ db_get [] (txtDb db) Nothing (encodeKey k)
    valToContentRec db k v1 v2

-- Odczyt i zapis całego dokumentu.

storeInDB :: Xdbs -> Document a -> IO ()
storeInDB db (Document _ _ e _) = do
    db_truncate [] (navDb db) Nothing
    db_truncate [] (txtDb db) Nothing
    db_truncate [] (tagDb db) Nothing
    db_truncate [] (tagIdx db) Nothing
    db_truncate [] (parentIdx db) Nothing
    db_truncate [] (txtIdx db) Nothing
    db_put_append [] (navDb db) Nothing (encodeNav $ NavData 9 (maxBound-2) 1 0)
    db_put_append [] (txtDb db) Nothing BS.empty
    db_put_append [] (tagDb db) Nothing BS.empty
    db_put [] (tagIdx db) Nothing BS.empty (encodeKey 1)
    let marked = markPrePost e
    storeContentInDB db 1 marked
    db_compact [] (parentIdx db) Nothing Nothing Nothing
    return ()

retrieveFromDB :: Xdbs -> IO (Document Mark)
retrieveFromDB db = do
    e <- retrieveContentFromDB db 2
    return (Document (Prolog Nothing [] Nothing []) [] e [])

-- Odpytywanie o potomków.

addGuard = maybe (encodeKey maxBound, undefined) id

retrieveDescendants :: Content Mark -> IO [Content Mark]
retrieveDescendants (Elem _ (PrePost (Just db) pre size)) 
        | size == 1 = return []
        | otherwise = unsafeInterleaveIO $ do
    dbc <- db_cursor [] (navDb db) Nothing
    Just v1 <- dbc_get_sz (Just 13) [] dbc (encodeKey $ pre + 1)
    ~(Just v2) <- unsafeInterleaveIO $ db_get [] (txtDb db) Nothing (encodeKey $ pre + 1)
    c1 <- valToContentRec db (pre + 1) v1 v2
    cs <- getRest dbc
    return (c1:cs)
    where
    getRest dbc = unsafeInterleaveIO $ do
        (k, v1) <- liftM addGuard $ dbc_move_get_sz (Just 4) (Just 13) [] dbc DB_NEXT
        ~(Just v2) <- unsafeInterleaveIO $ db_get [] (txtDb db) Nothing k
        if decodeKey k >= pre + size then dbc_close dbc >> return [] else do
        c <- valToContentRec db (decodeKey k) v1 v2
        cs <- getRest dbc
        return (c:cs)

retrieveChildren :: Content Mark -> IO [Content Mark]
retrieveChildren (Elem _ (PrePost (Just db) pre size)) 
        | size == 1 = return []
        | otherwise = unsafeInterleaveIO $ do
    dbc <- db_cursor [] (navDb db) Nothing
    getRest dbc (pre + 1)
    where
    getRest dbc i | i >= pre + size = dbc_close dbc >> return []
                  | otherwise = unsafeInterleaveIO $ do
        q <- dbc_get_sz (Just 13) [] dbc (encodeKey i)
        case q of
            Nothing -> dbc_close dbc >> return []
            Just v1 -> do
                ~(Just v2) <- unsafeInterleaveIO $ db_get [] (txtDb db) Nothing (encodeKey i)
                c <- valToContentRec db i v1 v2
                let (PrePost _ _ s) = info c
                cs <- getRest dbc (i + s)
                return (c:cs)

